# -*- coding: utf-8 -*-

"""Collection of default imports bundle."""

from .common_utils import *
from .detector_utils import *
from .jetbrains_utils import *
from .vscode_utils import *
from .installers import *



# -*- coding: utf-8 -*-

# pylint: disable=invalid-name

# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""Styles for the application."""
from __future__ import annotations

import ast
import contextlib
import os
import re
import string
import typing

from qtpy.QtCore import QSize  # pylint: disable=no-name-in-module
from qtpy.QtGui import QColor, QIcon

from anaconda_navigator.config import CONF
from anaconda_navigator.static import images
from anaconda_navigator.static.css import GLOBAL_SASS_STYLES_PATH, LIGHT_MODE_STYLES_PATH, DARK_MODE_STYLES_PATH

BLUR_SIZE = 10


class SassVariables:  # pylint: disable=too-few-public-methods,too-many-instance-attributes
    """Enum to hold SASS defined variables."""

    def __init__(self) -> None:
        """Enum to hold SASS defined variables."""
        self.SHADOW_BLUR_RADIUS = 7  # Used for dialogs
        self.WIDGET_APPLICATION_TOTAL_HEIGHT = 200
        self.WIDGET_APPLICATION_TOTAL_WIDTH = 200
        self.WIDGET_CONTENT_PADDING = 5
        self.WIDGET_CONTENT_TOTAL_HEIGHT = 200
        self.WIDGET_CONTENT_TOTAL_WIDTH = 200
        self.WIDGET_CONTENT_PADDING = 5
        self.WIDGET_CONTENT_MARGIN = 5
        self.WIDGET_ENVIRONMENT_TOTAL_HEIGHT = 50
        self.WIDGET_IMPORT_ENVIRONMENT_TOTAL_HEIGHT = 55
        self.WIDGET_ENVIRONMENT_TOTAL_WIDTH = 25
        self.WIDGET_APPLICATION_TOTAL_WIDTH = 260
        self.WIDGET_APPLICATION_TOTAL_HEIGHT = 295
        self.WIDGET_CHANNEL_DIALOG_WIDTH = 400
        self.WIDGET_CHANNEL_TOTAL_WIDTH = 300
        self.WIDGET_CHANNEL_TOTAL_HEIGHT = 40
        self.WIDGET_CHANNEL_PADDING = 5
        self.WIDGET_RUNNING_APPS_WIDTH = 450
        self.WIDGET_RUNNING_APPS_TOTAL_WIDTH = 350
        self.WIDGET_RUNNING_APPS_TOTAL_HEIGHT = 55
        self.WIDGET_RUNNING_APPS_PADDING = 10
        self.WIDGET_LOGIN_CARD_TOTAL_WIDTH = 315
        self.WIDGET_LOGIN_CARD_TOTAL_HEIGHT = 200

        self.ICON_ACTION_NOT_INSTALLED = os.path.join(images.STYLED_ICONS_PATH, 'check-box-blank.svg')
        self.ICON_ACTION_INSTALLED = os.path.join(images.IMAGE_PATH, 'icons', 'check-box-checked-active.svg')
        self.ICON_ACTION_REMOVE = os.path.join(images.IMAGE_PATH, 'icons', 'mark-remove.svg')
        self.ICON_ACTION_ADD = os.path.join(images.IMAGE_PATH, 'icons', 'mark-install.svg')
        self.ICON_ACTION_UPGRADE = os.path.join(images.IMAGE_PATH, 'icons', 'mark-upgrade.svg')
        self.ICON_ACTION_DOWNGRADE = os.path.join(images.IMAGE_PATH, 'icons', 'mark-downgrade.svg')
        self.ICON_UPGRADE_ARROW = os.path.join(images.IMAGE_PATH, 'icons', 'update-app-active.svg')
        self.ICON_SPACER = os.path.join(images.IMAGE_PATH, 'conda-manager-spacer.svg')
        self.ICON_PYTHON = os.path.join(images.IMAGE_PATH, 'python-logo.svg')
        self.ICON_ANACONDA = os.path.join(images.IMAGE_PATH, 'anaconda-logo.svg')

        self.COLOR_FOREGROUND_NOT_INSTALLED = '#666'
        self.COLOR_FOREGROUND_UPGRADE = '#00A3E0'
        self.SIZE_ICONS = (32, 32)
        self.STYLED_ICONS_COLOR = '#FFFFFF'

    def process_palette(self) -> dict[str, QIcon | QColor | QSize]:
        """Turn the styles _palette into QIcons or QColors for use in the model."""
        palette: dict[str, QIcon | QColor | QSize] = {}

        for key in dir(self):
            item: QIcon | QColor | QSize | None = None

            if key.startswith('ICON_'):
                item = QIcon(getattr(self, key))
            elif key.startswith('COLOR_'):
                item = QColor(getattr(self, key))
            elif key.startswith('SIZE_'):
                item = QSize(*getattr(self, key))

            if item:
                palette[key] = item

        return palette

    def __repr__(self):
        """Return a pretty formatted representation of the enum."""
        keys = []
        representation = 'SASS variables enum: \n'
        for key in self.__dict__:
            if key[0] in string.ascii_uppercase:
                keys.append(key)

        for key in sorted(keys):
            representation += f'    {key} = {self.__dict__[key]}\n'
        return representation


SASS_VARIABLES = SassVariables()


def flip_icons_color() -> None:
    """Change `fill` of svg icons according to current color pallete."""
    replace_pair: typing.Final[tuple[str, str]] = ('$STYLE_MODE_ICON_COLOR', SASS_VARIABLES.STYLED_ICONS_COLOR)

    for icon_path in os.listdir(images.STYLED_ICON_TEMPLATES_PATH):
        if not icon_path.endswith('.svg'):
            continue

        with open(os.path.join(images.STYLED_ICON_TEMPLATES_PATH, icon_path), 'r', encoding='utf-8') as f:
            svg_icon = f.read().replace(*replace_pair)

        with open(os.path.join(images.STYLED_ICONS_PATH, icon_path), 'w', encoding='utf-8') as f:
            f.write(svg_icon)


def load_sass_variables(path: str) -> SassVariables:
    """Parse Sass file styles and get custom values for used in code."""
    global SASS_VARIABLES  # pylint: disable=global-statement
    SASS_VARIABLES = SassVariables()

    with open(path, 'rt', encoding='utf-8') as f:
        data = f.read()

    pattern = re.compile(r'[$]\S*:.*?;')
    variables = re.findall(pattern, data)
    for var in variables:
        name, value = var[1:-1].split(':')
        if name[0] in string.ascii_uppercase:
            value = value.strip()
            with contextlib.suppress(BaseException):
                value = ast.literal_eval(value)
            setattr(SASS_VARIABLES, name, value)
    return SASS_VARIABLES


def load_style_sheet() -> str:
    """Load css styles file and parse to include custom variables."""
    load_sass_variables(GLOBAL_SASS_STYLES_PATH)
    styles_path = DARK_MODE_STYLES_PATH if CONF.get('main', 'dark_mode', False) else LIGHT_MODE_STYLES_PATH

    with open(styles_path, 'rt', encoding='utf-8') as f:
        data = f.read()

    load_sass_variables(styles_path.replace('.css', '.scss'))
    flip_icons_color()

    styled_image_path = images.DARK_IMAGE_PATH if CONF.get('main', 'dark_mode', False) else images.LIGHT_IMAGE_PATH
    data = data.replace(
        '$IMAGE_PATH', images.IMAGE_PATH.replace('\\', '/') if os.name == 'nt' else images.IMAGE_PATH
    ).replace(
        '$STYLED_IMAGE_PATH', styled_image_path.replace('\\', '/')
        if os.name == 'nt' else styled_image_path
    ).replace(
        '$STYLED_ICONS_PATH', images.STYLED_ICONS_PATH.replace('\\', '/')
        if os.name == 'nt' else images.STYLED_ICONS_PATH
    )

    return data



# -*- coding: utf8 -*-

"""Utilities to manage login workflows and states."""

from __future__ import annotations

__all__ = ['LoginManager', 'LOGIN_MANAGER']

import typing

from qtpy import QtCore

from anaconda_navigator import config
from anaconda_navigator.utils import singletons
from anaconda_navigator.utils import telemetry
from anaconda_navigator.utils import workers
from anaconda_navigator.utils.logs import logger
from . import api


class LoginManager(QtCore.QObject):
    """Manage .cloud login flow."""

    sig_login_started = QtCore.Signal()
    sig_login_succeeded = QtCore.Signal(workers.TaskResult)
    sig_login_canceled = QtCore.Signal(workers.TaskResult)
    sig_login_failed = QtCore.Signal(workers.TaskResult)
    sig_login_done = QtCore.Signal(workers.TaskResult)

    sig_logout = QtCore.Signal()

    sig_state_changed = QtCore.Signal(bool)

    sig_throttle_started = QtCore.Signal()
    sig_throttle_finished = QtCore.Signal()

    def __init__(self, parent: QtCore.QObject | None = None) -> None:
        """Initialize new instance of a :class:`~CloudLoginManager`."""
        super().__init__(parent=parent)

        self.__pending: bool = False
        self.__worker: workers.TaskWorker | None = None

        self.__throttle: bool = False

        self.__throttle_timer: typing.Final[QtCore.QTimer] = QtCore.QTimer(self)
        self.__throttle_timer.setInterval(5_000)  # 5 seconds
        self.__throttle_timer.setSingleShot(True)
        self.__throttle_timer.timeout.connect(self.__stop_throttle)

    # login

    def login(self, origin: str = 'unknown') -> None:
        """Start a login process."""
        if self.__throttle:
            logger.debug('cloud login throttled')
            return
        logger.debug('cloud login initiated')

        telemetry.ANALYTICS.instance.event('cloud-login-requested', {'origin': origin})
        if self.__worker is not None:
            self.__pending = True
            self.__worker.cancel()
        else:
            self.__login()

    def __login(self) -> None:
        """Initialize a worker to process login workflow."""
        self.__start_throttle()

        self.__worker = typing.cast(workers.TaskWorker, api.CloudAPI().login.worker())
        self.__worker.signals.sig_start.connect(self.sig_login_started)
        self.__worker.signals.sig_succeeded.connect(self.sig_login_succeeded)
        self.__worker.signals.sig_canceled.connect(self.sig_login_canceled)
        self.__worker.signals.sig_failed.connect(self.sig_login_failed)
        self.__worker.signals.sig_done.connect(self.__login_done)
        self.__worker.start()

    def __login_done(self, result: workers.TaskResult) -> None:
        """Process result of a login processing."""
        self.__stop_throttle()
        self.__worker = None
        self.sig_login_done.emit(result)

        if result.status == workers.TaskStatus.SUCCEEDED:
            logger.debug('cloud login succeeded')
            telemetry.ANALYTICS.instance.event(
                'login_changed', {
                    'action': telemetry.AccountAction.LOGIN,
                    'edition': config.AnacondaBrand.CLOUD,
                }, user_properties={'cloud-edition': config.AnacondaBrand.CLOUD})
            self.sig_state_changed.emit(True)
        elif result.status == workers.TaskStatus.CANCELED:
            if self.__pending:
                self.__pending = False
                self.__login()

            logger.debug('cloud login canceled')
            telemetry.ANALYTICS.instance.event('cancel-cloud-login')
        elif result.status == workers.TaskStatus.FAILED:
            logger.debug('cloud login failed')
            telemetry.ANALYTICS.instance.event('cloud-login-failed')
        else:  # just in case we add new status and forget to add it here
            logger.debug('cloud login done')  # type: ignore

    # logout

    def logout(self) -> None:
        """Log out from .cloud account."""
        if not api.CloudAPI().token:
            logger.debug('cloud logout ignored')
            return
        logger.debug('cloud logout initiated')

        if self.__worker is not None:
            self.__worker.cancel()

        api.CloudAPI().logout()
        telemetry.ANALYTICS.instance.event(
            'login_changed', {
                'action': telemetry.AccountAction.LOGOUT,
                'edition': config.AnacondaBrand.CLOUD,
            },
            user_properties={'cloud-edition': '-'})
        self.sig_logout.emit()
        self.sig_state_changed.emit(False)
        logger.debug('cloud logout succeeded')

    # throttle

    def __start_throttle(self) -> None:
        """
        Start login process throttling.

        It doesn't allow to spam logins by giving each login attempt 5 seconds to start.
        """
        if self.__throttle:
            return
        logger.debug('cloud login throttling started')
        self.__throttle = True
        self.__throttle_timer.start()
        self.sig_throttle_started.emit()

    def __stop_throttle(self) -> None:
        """Finish login throttling."""
        if not self.__throttle:
            return
        logger.debug('cloud login throttling finished')

        self.__throttle_timer.stop()
        self.__throttle = False
        self.sig_throttle_finished.emit()


LOGIN_MANAGER: typing.Final[singletons.Singleton[LoginManager]] = singletons.SingleInstanceOf(LoginManager)



# -*- coding: utf-8 -*-

"""Solvers for Conda configuration entries, that are not valid."""

from __future__ import annotations

__all__ = ()

import html
import typing

from anaconda_navigator.utils import configurations
from anaconda_navigator.utils import notifications
from . import core

if typing.TYPE_CHECKING:
    from .. import types as conda_types


@core.POOL.register_function(filters={'exception_name': 'CustomValidationError', 'parameter_name': 'ssl_verify'})
def solve_ssl_verify(error: 'conda_types.CondaValidationErrorOutput') -> notifications.Notification:
    """
    Solve issues with invalid `ssl_verify` preference.

    Solution: remove invalid values.
    """
    updater: configurations.Updater
    with configurations.YamlUpdater(path=error['source']) as updater:
        message: str = 'SSL certificate(s) are no longer available'

        first: bool = True
        value: str | None
        for value in (
                updater.content.pop('ssl_verify', None),
                updater.content.pop('verify_ssl', None),
        ):
            if value is not None:
                if first:
                    message += ':'
                    first = False
                message += f'<br>- {html.escape(value)}'

        return notifications.Notification(
            message=message,
            caption='Broken Conda configuration',
            tags=('conda', 'ssl_certificate'),
        )



# -*- coding: utf-8 -*-

"""Helper utilities and shortcuts for the :mod:`~anaconda_navigator.utils.workers`."""

from __future__ import annotations

__all__ = ['parallel_reduce', 'parallel_collect']

import typing

from . import workers


T = typing.TypeVar('T')


EMPTY: typing.Any = object()


def parallel_reduce(
        function: typing.Callable[[T, typing.Any], T],
        tasks: typing.Iterable[workers.TaskWorker],
        initial: T = EMPTY,
) -> T:
    """Run multiple tasks in parallel and reduce their results into a single value."""
    threads: typing.List[workers.TaskThread] = [task.thread() for task in tasks]
    threads.reverse()

    while threads:
        thread: workers.TaskThread = threads.pop()
        thread.wait()
        try:
            current: T = typing.cast(workers.TaskResult, thread.result).result
            if initial is EMPTY:
                initial = current
            else:
                initial = function(initial, current)
        except Exception:
            for thread in threads:
                thread.cancel()
            raise
    if initial is EMPTY:
        raise TypeError('at least one task or initial value must be provided')
    return initial


def list_append(collection: typing.List[T], value: T) -> typing.List[T]:
    """
    Append new value to the list.

    Returns the list to which value was appended.
    """
    collection.append(value)
    return collection


def parallel_collect(tasks: typing.Iterable[workers.TaskWorker]) -> typing.List[typing.Any]:
    """
    Run multiple tasks in parallel and return all their results.

    Results are returned in the same order tasks were provided.
    """
    return parallel_reduce(list_append, tasks, [])



# -*- coding: utf-8 -*-

"""Additional utility functions to use for installable applications."""

from __future__ import annotations

__all__ = ['install_launch', 'install_dmg']

import stat

import webbrowser
import collections.abc
import contextlib
import hashlib
import io
import os
import posixpath
import shutil
import subprocess  # nosec: B404
import sys
import tempfile
import typing
import urllib.parse

import requests

from anaconda_navigator.config import preferences
from anaconda_navigator.static.content import DATA_PATH
from anaconda_navigator.utils.logs import logger

if typing.TYPE_CHECKING:
    from .. import base


def safe_domain(domain: str) -> bool:
    """Check if installer at :code:`domain` is safe to use."""
    if preferences.ALLOW_UNSAFE_INSTALLERS:
        return True
    domain = f'.{domain}'
    return any(domain.endswith(item) for item in preferences.SAFE_INSTALLER_DOMAINS)


def safe_filename(filename: str) -> str | None:
    """Check if installer at :code:`filename` is safe to use."""
    if not filename:
        return None
    filename = os.path.abspath(os.path.join(DATA_PATH, filename))

    if preferences.ALLOW_UNSAFE_INSTALLERS:
        return filename

    try:
        if os.path.samefile(os.path.commonpath([DATA_PATH, filename]), DATA_PATH):
            return filename
    except ValueError:
        return None

    return None


@contextlib.contextmanager
def retrieve(url: str) -> collections.abc.Iterator[str]:
    """
    Download installer.

    Must be used as a context manager.

    :param url: URI of the installer (might be a path to local file).
    :return: Path to a downloaded file.
    """
    filename: str | None
    parsed: urllib.parse.SplitResult = urllib.parse.urlsplit(url)
    if (parsed.scheme in ('', 'file')) and (filename := safe_filename(parsed.path)):
        yield filename
    elif safe_domain(parsed.hostname or ''):
        response: requests.Response
        with requests.get(url, stream=True, timeout=30.0, verify=True) as response:
            response.raise_for_status()

            descriptor: int
            stream: io.BufferedWriter
            descriptor, filename = tempfile.mkstemp(suffix=posixpath.splitext(parsed.path)[1])
            with os.fdopen(descriptor, 'wb') as stream:
                chunk: bytes
                for chunk in response.iter_content(chunk_size=4_194_304):
                    stream.write(chunk)

        try:
            yield filename
        finally:
            try:
                os.remove(filename)
            except OSError:
                pass
    else:
        raise ValueError('source is unsafe')


@contextlib.contextmanager
def retrieve_and_validate(url: str, **kwargs: str | None) -> collections.abc.Iterator[str]:
    """Same as :func:`~retrieve`, but also checks for resulting files hash sums."""
    filename: str
    with retrieve(url) as filename:
        query: list[tuple[hashlib._Hash, str]] = [
            (hashlib.new(name), value)
            for name, value in kwargs.items()
            if value
        ]
        if query:
            expectation: str
            hasher: hashlib._Hash
            stream: io.BufferedReader
            with open(filename, 'rb') as stream:
                buffer: bytes
                while buffer := stream.read(4_294_967_296):
                    for hasher, _ in query:
                        hasher.update(buffer)
            for hasher, expectation in query:
                if hasher.hexdigest() != expectation:
                    raise ValueError(
                        f'installer hash missmatch for {os.path.basename(filename)}: '
                        f'got {hasher.hexdigest()!r} instead of {expectation!r}'
                    )
        yield filename


def install_launch(  # pylint: disable=too-many-arguments
        app: 'base.BaseInstallableApp',  # pylint: disable=unused-argument
        *,
        installer_url: str,
        fallback_url: str,
        flags: collections.abc.Sequence[str] = (),
        md5: str | None = None,
        sha256: str | None = None,
) -> None:
    """
    Install application by launching an installer.

    :param installer_url: URL to installer.
    :param fallback_url: URL to open in browser if installer fails.
    :param flags: Extra command line flags to pass to the installer.
    :param md5: Expected MD5 hash of the installer file.
    :param sha256: Expected SHA256 hash of the installer file.
    """
    try:
        with retrieve_and_validate(installer_url, md5=md5, sha256=sha256) as filename:
            if sys.platform in ('linux', 'darwin'):
                info: os.stat_result = os.stat(filename)
                mode: int = info.st_mode | stat.S_IXUSR | stat.S_IXGRP
                if mode != info.st_mode:
                    os.chmod(filename, mode)
            subprocess.check_call([filename, *flags])  # nosec: B603
    except Exception:  # pylint: disable=broad-exception-caught
        webbrowser.open(fallback_url)


def install_dmg(  # pylint: disable=too-many-arguments
        app: 'base.BaseInstallableApp',  # pylint: disable=unused-argument
        *,
        installer_url: str,
        fallback_url: str,
        md5: str | None = None,
        reinstall: bool = False,
        sha256: str | None = None,
        system: bool = True,
) -> None:
    """
    Install application from a DMG installer.

    .. warning::

        Works only on macOS.

    :param installer_url: URL to installer.
    :param fallback_url: URL to open in browser if installer fails.
    :param md5: Expected MD5 hash of the installer file.
    :param reinstall: Reinstall application if it is already installed.
    :param sha256: Expected SHA256 hash of the installer file.
    :param system: Install application system-wide instead of "only for user".
    """
    try:
        import dmglib  # pylint: disable=import-error,import-outside-toplevel

        filename: str
        with retrieve_and_validate(installer_url, md5=md5, sha256=sha256) as filename:
            if dmglib.dmg_already_attached(filename):
                dmglib.dmg_detach_already_attached(filename)

            image: dmglib.DiskImage = dmglib.DiskImage(filename)
            try:
                if image.has_license_agreement():
                    logger.debug('DMG %r requires user to accept license agreement manually', installer_url)
                    raise ValueError('installer requires user to accept license agreement')

                if system:
                    target: str = '/Applications'
                else:
                    target = os.path.join(os.environ['HOME'], 'Applications')

                applications: list[tuple[str, str]] = [
                    (name, os.path.join(root, name))
                    for root in image.attach()
                    for name in os.listdir(root)
                    if name.endswith('.app')
                ]
                if not applications:
                    logger.debug('DMG %r does not contain any applications', installer_url)
                    raise ValueError('Installer does not contain any applications')

                app_name: str
                app_root: str
                for app_name, app_root in applications:
                    app_target: str = os.path.join(target, app_name)
                    if os.path.exists(app_target):
                        if reinstall:
                            shutil.rmtree(app_target)
                        else:
                            continue
                    shutil.copytree(app_root, app_target, symlinks=True)
            finally:
                image.detach()
    except Exception:  # pylint: disable=broad-exception-caught
        webbrowser.open(fallback_url)



# -*- coding: utf-8 -*-

"""Core structures for third-party application descriptions."""

from __future__ import annotations

__all__ = ['AppPatch', 'BaseApp', 'BaseWebApp', 'BaseInstallableApp']

import abc
import collections.abc
import os
import re
import types
import typing
import webbrowser

import attr

from anaconda_navigator.static.images import ANACONDA_ICON_256_PATH, APPLICATIONS_PATH
from anaconda_navigator.utils import constants

from . import detectors
from . import exceptions
from . import import_utils
from . import parsing_utils
from . import validation


if typing.TYPE_CHECKING:
    from anaconda_navigator.api import process
    from anaconda_navigator.api import types as api_types
    from anaconda_navigator.config import user as user_config


AppPatchField = typing.Literal['display_name', 'description', 'image_path', 'rank']


class AppPatchDelta(typing.TypedDict, total=False):
    """Change provided by :class:`~AppPatch`."""
    display_name: str
    description: str
    image_path: str
    rank: int


APP_PATCH_FIELD_KEYS: 'typing.Final[typing.Tuple[AppPatchField, ...]]' = AppPatchField.__args__  # type: ignore


@attr.s(auto_attribs=True, cache_hash=True, frozen=True, order=False, slots=True)
class AppPatch:
    """Patch for the application defined outside this component."""

    app_name: str = attr.ib(default='')
    display_name: typing.Optional[str] = attr.ib(default=None)
    description: typing.Optional[str] = attr.ib(default=None)
    image_path: typing.Optional[str] = attr.ib(default=None)
    is_available: bool = attr.ib(default=True)
    rank: typing.Optional[int] = attr.ib(default=None)

    @property
    def delta(self) -> 'AppPatchDelta':  # noqa: D401
        """Dict-like content of the patch."""
        result: 'AppPatchDelta' = {}

        # NOTE: should be replaced with py38+: fields = AppPatchField.__args__

        field: 'AppPatchField'
        for field in APP_PATCH_FIELD_KEYS:
            current: typing.Any = getattr(self, field)
            if current is not None:
                result[field] = current

        return result

    @typing.overload
    def apply_to(self, instance: AppPatch) -> AppPatch:
        """Modify other AppPatch."""

    @typing.overload
    def apply_to(self, instance: 'api_types.Application') -> 'api_types.Application':
        """Apply all values from a patch to an application description dictionary."""

    def apply_to(
            self,
            instance: typing.Union[AppPatch, 'api_types.Application'],
    ) -> typing.Union[AppPatch, 'api_types.Application']:
        """Apply patch to another instance."""
        if isinstance(instance, AppPatch):
            return attr.evolve(
                instance,
                is_available=self.is_available,
                **self.delta,
            )

        instance.update(typing.cast('api_types.Application', self.delta))
        return instance


def _call(function: str) -> typing.Any:
    """Execute imported function and return the execution result."""
    func: typing.Callable[[], typing.Any] = import_utils.import_callable(function)
    with validation.catch_exception():
        return func()


def _image_path(value: str) -> str:
    """Normalize `image_path` value."""
    result: str = os.path.join(APPLICATIONS_PATH, value)
    if os.path.isfile(result):
        return result
    return value


@attr.s(auto_attribs=True, cache_hash=True, frozen=True, order=False)
class BaseApp(metaclass=abc.ABCMeta):
    """
    Root description of the application.

    :param app_type: Type of the application (web application, installable, etc.).
    :param app_name: Alias of the application in a package-naming format.
    :param display_name: Name of the application to show in home tile.
    :param description: Description of the application to show in home tile.
    :param image_path: Application icon to show in home tile.
    :param config: Application configuration.
    """

    __app_types__: typing.ClassVar[typing.MutableMapping[str, typing.Type['BaseApp']]] = {}

    app_type: constants.AppType
    app_name: str
    display_name: str
    description: str
    image_path: str
    non_conda: bool
    config: 'user_config.UserConfig'
    is_available: bool
    rank: int

    def __init_subclass__(
            cls,  # pylint: disable=unused-argument
            alias: typing.Optional[constants.AppType] = None,
            **kwargs: typing.Any,
    ) -> None:
        """
        Initialize new :class:`~BaseApp` subclass.

        Used to register new classes in registry.
        """
        if alias is None:
            return

        if alias in cls.__app_types__:
            raise ValueError(f'{alias!r} is already registered')

        cls.__app_types__[alias] = cls

    @property
    def tile_definition(self) -> 'api_types.Application':  # noqa: D401
        """Definition of the application tile."""
        return {
            'app_type': self.app_type,
            'description': self.description,
            'display_name': self.display_name,
            'image_path': self.image_path,
            'name': self.app_name,
            'non_conda': self.non_conda,
            'rank': self.rank,
        }

    @classmethod
    def parse_configuration(
            cls,
            context: parsing_utils.ParsingContext,
            configuration: typing.Mapping[str, typing.Any],
            app_name: str,
    ) -> typing.Union[None, 'BaseApp', 'AppPatch']:
        """Generate applications according to the `configuration`."""
        validation.OfType(typing.Mapping)(configuration)
        mirror: typing.Dict[str, typing.Any] = dict(configuration)

        app_type: str
        target_cls: typing.Optional[typing.Type[BaseApp]]
        display_name: typing.Optional[str]
        description: typing.Optional[str]
        image_path: typing.Optional[str]
        rank: typing.Optional[int]
        try:
            app_type = validation.ItemPopper('app_type')(mirror)
        except exceptions.ValidationError:
            target_cls = None
            display_name = validation.ItemPopper('display_name', default=None).convert(validation.IS_OPT_STR)(mirror)
            description = validation.ItemPopper('description', default=None).convert(validation.IS_OPT_STR)(mirror)
            image_path = validation.ItemPopper('image_path', default=None).convert(
                validation.IS_NONE | validation.IS_STR.convert(_image_path),
            )(mirror)
            rank = validation.ItemPopper('rank', default=None).convert(validation.IS_OPT_INT)(mirror)
        else:
            target_cls = cls.__app_types__[validation.OfChoices(*cls.__app_types__)(app_type, field='app_type')]
            display_name = validation.ItemPopper('display_name').convert(validation.IS_STR)(mirror)
            description = validation.ItemPopper('description', default='').convert(validation.IS_STR)(mirror)
            image_path = validation.ItemPopper('image_path', default=ANACONDA_ICON_256_PATH).convert(
                validation.IS_STR.convert(_image_path),
            )(mirror)
            rank = validation.ItemPopper('rank', default=0).convert(validation.IS_INT)(mirror)

        is_available: bool = validation.ItemPopper('is_available', default=True).convert(
            validation.sequence_of(str, bool),
            validation.FindValue(
                validation.OfType(bool) | validation.OfType(str).convert(_call).convert(validation.IS_BOOL),
            ),
        )(mirror)

        if target_cls is None:
            validation.IsEmptyMapping()(mirror)
            return AppPatch(
                app_name=app_name,
                display_name=display_name,
                description=description,
                image_path=image_path,
                is_available=is_available,
                rank=rank,
            )

        if not is_available:
            return None

        return target_cls._parse_configuration(  # pylint: disable=protected-access
            context=context,
            configuration=mirror,

            app_name=app_name,
            config=context.user_configuration,
            display_name=display_name,
            description=description,
            image_path=image_path,
            is_available=is_available,
            rank=rank,
        )

    @staticmethod
    @abc.abstractmethod
    def _parse_configuration(
            context: parsing_utils.ParsingContext,
            configuration: typing.MutableMapping[str, typing.Any],
            **kwargs: typing.Any,  # app_name, config, display_name, description, image_path, is_available, rank
    ) -> BaseApp:
        """Parse configuration for this particular :class:`~BaseApp`."""


@attr.s(auto_attribs=True, cache_hash=True, collect_by_mro=True, frozen=True, order=False)
class BaseWebApp(BaseApp, alias=constants.AppType.WEB):
    """
    Common base for all external web applications.

    Most of the params reused from the :class:`~BaseApp`.

    :param url: Web page to open for the application.
    """

    app_type: constants.AppType = attr.ib(default=constants.AppType.WEB, init=False)
    non_conda: bool = attr.ib(default=True, init=False)

    url: str

    @property
    def tile_definition(self) -> 'api_types.Application':  # noqa: D401
        """Definition of the application tile."""
        result: 'api_types.Application' = super().tile_definition
        result['installed'] = True
        return result

    @classmethod
    def _parse_configuration(
            cls,
            context: parsing_utils.ParsingContext,
            configuration: typing.MutableMapping[str, typing.Any],
            **kwargs: typing.Any,  # app_name, config, display_name, description, image_path, is_available, rank
    ) -> BaseWebApp:
        """Parse configuration for this particular :class:`~BaseApp`."""
        url: str = validation.ItemPopper('url').convert(validation.IS_STR)(configuration)
        validation.IsEmptyMapping()(configuration)

        return BaseWebApp(url=url, **kwargs)


InstallFunction = typing.Callable[['BaseInstallableApp'], None]
InstallExtensionsFunction = typing.Callable[['BaseInstallableApp'], 'process.ProcessWorker']
UpdateConfigFunction = typing.Callable[['BaseInstallableApp', str], None]


def _install(value: str | collections.abc.Mapping[str, typing.Any]) -> InstallFunction | None:
    """Validate and resolve value of `install`."""
    if isinstance(value, collections.abc.Mapping):
        value = dict(value)

        name: str = validation.ItemPopper('func').convert(validation.IS_STR)(value)
        function: collections.abc.Callable[..., typing.Any] = import_utils.import_callable(name)

        arguments: collections.abc.Sequence[typing.Any] = validation.ItemPopper('args', default=()).convert(
            validation.IS_SEQ,
        )(value)

        def install(self: 'BaseInstallableApp') -> None:
            """Install application"""
            return function(self, *arguments, **value)

        return install

    if re.match(r'^[a-z][a-z0-9.+-]*://', value):
        def install(self: 'BaseInstallableApp') -> None:  # pylint: disable=function-redefined,unused-argument
            """Install application"""
            webbrowser.open(value)

        return install

    return import_utils.import_callable(value)


def to_string_tuple(value: typing.Iterable[typing.Any]) -> typing.Sequence[str]:
    """Convert any iterable value to a string tuple."""
    if isinstance(value, str):
        return (value,)
    return tuple(map(str, value))


@attr.s(auto_attribs=True, cache_hash=True, collect_by_mro=True, frozen=True, order=False)
class BaseInstallableApp(BaseApp, alias=constants.AppType.INSTALLABLE):
    """
    Common base for all external installable applications.

    Most of the params reused from the :class:`~BaseApp`.

    :param detector: Detector of the application location.
    :param extra_arguments: Additional arguments to pass to the application.
    :param process_api: API to spawn new processes (for extensions installation etc.).
    """

    install: typing.ClassVar[InstallFunction]

    app_type: constants.AppType = attr.ib(default=constants.AppType.INSTALLABLE, init=False)
    non_conda: bool = attr.ib(default=True, init=False)

    extra_arguments: typing.Sequence[typing.Any] = attr.ib(converter=to_string_tuple)
    _detector: 'detectors.Source'
    _process_api: process.WorkerManager
    _location: typing.Optional[detectors.DetectedApplication] = attr.ib(default=None, init=False)

    def __attrs_post_init__(self) -> None:
        """Additional initialization of the class."""
        context: detectors.DetectorContext = detectors.DetectorContext(
            app_name=self.app_name,
            user_configuration=self.config,
        )
        primary_location: 'typing.Optional[detectors.DetectedApplication]'
        secondary_location: 'typing.Optional[detectors.DetectedApplication]' = None
        for primary_location in self._detector(context=context):
            if primary_location.complete:
                break
            if secondary_location is not None:
                continue
            primary_location = primary_location.replace(version='')
            if primary_location.complete:
                secondary_location = primary_location
        else:
            primary_location = secondary_location
        object.__setattr__(self, '_location', primary_location)

        self.config.set('applications', f'{self.app_name}_path', self.root or '')

    @property
    def tile_definition(self) -> 'api_types.Application':  # noqa: D401
        """Definition of the application tile."""
        result: 'api_types.Application' = super().tile_definition
        result['command'] = self.executable or ''
        result['extra_arguments'] = self.extra_arguments
        result['installed'] = self.is_installed
        result['versions'] = self.versions
        return result

    @property
    def executable(self) -> typing.Optional[str]:  # noqa: D401
        """Command to execute the application."""
        if self._location is None:
            return None

        from anaconda_navigator.utils import launch  # pylint: disable=import-outside-toplevel

        return launch.safe_argument(self._location.executable)

    @property
    def is_installation_enabled(self) -> bool:  # noqa: D401
        """Application can be installed."""
        return hasattr(self, 'install')

    @property
    def is_installed(self) -> bool:  # noqa: D401
        """Application is installed in local system."""
        return bool(self.executable)

    @property
    def root(self) -> typing.Optional[str]:  # noqa: D401
        """Directory, in which application is installed."""
        if self._location is None:
            return None
        return self._location.root

    @property
    def version(self) -> typing.Optional[str]:  # noqa: D401
        """Version of the installed application."""
        if self._location is None:
            return None
        return self._location.version

    @property
    def versions(self) -> typing.Sequence[str]:  # noqa: D401
        """List of available application versions."""
        if self._location is None:
            return []
        return [self._location.version]

    @classmethod
    def _parse_configuration(
            cls,
            context: parsing_utils.ParsingContext,
            configuration: typing.MutableMapping[str, typing.Any],
            **kwargs: typing.Any,  # app_name, config, display_name, description, image_path, is_available, rank
    ) -> BaseInstallableApp:
        """Parse configuration for this particular :class:`~BaseApp`."""
        try:
            detector: detectors.Source = validation.ItemPopper('detector').convert(
                validation.IS_STR.convert(_call) | validation.IS_MAP.convert(detectors.Detector.parse_configuration),
                validation.OfType(detectors.Source),
            )(configuration)
        except exceptions.ValidationError:
            # show "install" option even if the application itself is unavailable
            # otherwise - remove `except` section
            detector = detectors.ZeroSource()

        extra_arguments: typing.Tuple[str, ...] = validation.ItemPopper('extra_arguments', default=None).convert(
            validation.IS_NONE.convert(lambda _: ()) | validation.IS_STR.convert(_call) | validation.IS_SEQ,
            validation.IS_SEQ,
            validation.EachValue(validation.IsAny().convert(str)).as_converter(tuple),
        )(configuration)

        result: BaseInstallableApp = BaseInstallableApp(
            detector=detector,
            process_api=context.process_api,
            extra_arguments=extra_arguments,
            **kwargs,
        )

        method: typing.Optional[typing.Callable[..., typing.Any]]
        method = validation.ItemPopper('install', default=None).convert(
            validation.sequence_of(str, bool, collections.abc.Mapping, None),
            validation.FindValue((validation.IS_STR | validation.IS_MAP).convert(_install), default=None),
        )(configuration)
        if method is not None:
            object.__setattr__(result, 'install', types.MethodType(method, result))

        method_name: str
        for method_name in ('install_extensions', 'update_config'):
            with validation.validation_to_warning(silent=True):
                method = validation.ItemPopper(method_name, default=None).convert(
                    validation.IS_NONE | validation.IS_STR.convert(import_utils.import_callable),
                )(configuration)
                if method is not None:
                    object.__setattr__(result, method_name, types.MethodType(method, result))

        validation.IsEmptyMapping()(configuration)

        return result

    def install_extensions(self) -> 'process.ProcessWorker':
        """Install app extensions."""
        return self._process_api.create_process_worker(['python', '--version'])

    def update_config(self, prefix: str) -> None:
        """Update user config to use selected Python prefix interpreter."""



"""Login reminder dialog."""

from __future__ import annotations

import typing

from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import QLabel, QVBoxLayout, QHBoxLayout

from anaconda_navigator.api import cloud
from anaconda_navigator.utils.widgets.signals import DialogOutcome, DialogValue
from anaconda_navigator.widgets import ButtonPrimary, ButtonLink, ButtonSecondaryTextual
from anaconda_navigator.widgets.dialogs.login import AccountOutcome, AccountValue
from anaconda_navigator.widgets.dialogs.login.reminder.styling import BaseCleanDialog
from anaconda_navigator.widgets.web.links import open_link

if typing.TYPE_CHECKING:
    from anaconda_navigator.widgets import main_window


class LoginReminderDialog(BaseCleanDialog):  # pylint: disable=too-many-instance-attributes
    """Login reminder dialog."""
    sig_dialog_request = Signal(DialogOutcome, DialogValue)
    sig_login_request = Signal(AccountOutcome, AccountValue)

    def __init__(self, parent: 'main_window.MainWindow', title: str = 'Sign in for AI Coding Help'):
        """Initialize new :class:`~LoginReminderDialog` instance."""
        super().__init__(parent, title)

        self.__image_label = QLabel()
        self.__image_label.setObjectName('LoginReminderImage')
        self.__image_label.setScaledContents(True)

        self.__message_label = QLabel('Sign in to Anaconda Cloud or create an account to access '
                                      'Anaconda Assistant in Jupyter Notebooks and JupyterLab.')
        self.__message_label.setWordWrap(True)

        self.__cloud_login_button = ButtonPrimary('Sign In')
        self.__cloud_login_button.setFocusPolicy(Qt.NoFocus)
        self.__repo_login_button = ButtonSecondaryTextual('Connect to a Repository')
        self.__repo_login_button.setFocusPolicy(Qt.NoFocus)

        self.__sign_up_label = QLabel('Don’t have an account? ')
        self.__sign_up_link = ButtonLink('Sign Up')
        self.__sign_up_link.setFocusPolicy(Qt.NoFocus)

        buttons_layout = QHBoxLayout()
        buttons_layout.setAlignment(Qt.AlignLeft)
        buttons_layout.addWidget(self.__cloud_login_button, alignment=Qt.AlignLeft)
        buttons_layout.addSpacing(16)
        buttons_layout.addWidget(self.__repo_login_button, alignment=Qt.AlignLeft)

        text_layout = QHBoxLayout()
        text_layout.setAlignment(Qt.AlignLeft)
        text_layout.addWidget(self.__sign_up_label, alignment=Qt.AlignLeft)
        text_layout.addSpacing(5)
        text_layout.addWidget(self.__sign_up_link, alignment=Qt.AlignLeft)

        self.__content_layout = QVBoxLayout()
        self.__content_layout.setAlignment(Qt.AlignTop)
        self.__content_layout.addWidget(self.__image_label, alignment=Qt.AlignLeft)
        self.__content_layout.addSpacing(24)
        self.__content_layout.addWidget(self.__message_label)
        self.__content_layout.addSpacing(20)
        self.__content_layout.addLayout(buttons_layout)
        self.__content_layout.addSpacing(24)

        self.__content_layout.addLayout(text_layout)

        self.__repo_login_button.clicked.connect(lambda: self.sig_dialog_request.emit(
            DialogOutcome.DIALOG_REQUEST, DialogValue.REPO_SELECTOR))

        self.__cloud_login_button.clicked.connect(lambda: self.sig_login_request.emit(
            AccountOutcome.LOGIN_REQUEST, AccountValue.CLOUD))
        cloud.LOGIN_MANAGER.instance.sig_login_done.connect(self.accept)

        self.__sign_up_link.clicked.connect(lambda: open_link('https://anaconda.cloud/sign-up'))

        self.sig_dialog_request.connect(self.reject)

        self.setLayout(self.__content_layout)

    def accept(self) -> None:
        """Override default `accept` to disconnect global signals."""
        cloud.LOGIN_MANAGER.instance.sig_login_done.disconnect(self.accept)
        return super().accept()



# -*- coding: utf-8 -*-

# pylint: disable=attribute-defined-outside-init,unused-argument

# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""Dialogs for importing environments."""

from __future__ import annotations

__all__ = ['ImportSelectorDialog']

import enum
import html
import os
import tempfile
import typing
import uuid

from qtpy import compat
from qtpy import QtCore
from qtpy import QtWidgets
import requests
import yaml

from anaconda_navigator.api import cloud
from anaconda_navigator.api.cloud import tools as cloud_tools
from anaconda_navigator.utils import styles
from anaconda_navigator.utils import telemetry
from anaconda_navigator.utils import workers
from anaconda_navigator import widgets
from anaconda_navigator.widgets import dialogs
from anaconda_navigator.widgets import common as global_common
from anaconda_navigator.widgets.lists import environments
from .. import utilities
from . import common
from . import multiaccount_dialogs

if typing.TYPE_CHECKING:
    from anaconda_navigator.widgets import main_window


DEFAULT_ERROR: typing.Final[str] = 'Unable to import environment due to an unknown error.'

ENVIRONMENT_FETCH_LIMIT: typing.Final[int] = 20
ENVIRONMENT_ORDER: typing.Final[cloud.EnvironmentSort] = cloud.EnvironmentSort.UPDATED_DESC


def env_to_spec(source: str, target: str, name: typing.Optional[str] = None) -> None:
    """
    Convert Conda environment file to Conda specification file.

    :param source: Path to the conda environment file.
    :param target: Path to the conda specification file.
    :param name: Name of the environment being converted.
    """
    content: typing.Any

    stream: typing.TextIO
    with open(source, 'rt', encoding='utf-8') as stream:
        content = yaml.safe_load(stream)

    content['name'] = name

    with open(target, 'wt', encoding='utf-8') as stream:
        yaml.dump(content, stream, default_flow_style=False)


def pip_to_spec(source: str, target: str, name: typing.Optional[str] = None) -> None:
    """
    Convert pip requirements file to Conda specification file.

    :param source: Path to the pip requirements file.
    :param target: Path to the conda specification file.
    :param name: Name of the environment being converted.
    """
    dependencies: typing.List[str]

    stream: typing.TextIO
    with open(source, 'rt', encoding='utf-8') as stream:
        dependencies = [
            line
            for line in map(str.strip, stream.read().splitlines())
            if line[:1] not in ('', '#')
        ]

    with open(target, 'wt', encoding='utf-8') as stream:
        yaml.dump(
            {
                'name': name,
                'dependencies': [
                    'python',
                    {'pip': dependencies},
                ],
            },
            stream,
            default_flow_style=False,
        )


class ImportSelectorType(enum.IntEnum):
    """
    Type of the file, imported from the file system.

    .. py:attribute:: NONE

        File is not yet selected.

    .. py:attribute:: ENVIRONMENT

        Conda environment file.

    .. py:attribute:: SPECIFICATION

        Conda specification file.

    .. py:attribute:: PIP

        Pip requirements file.
    """

    NONE = enum.auto()
    ENVIRONMENT = enum.auto()
    SPECIFICATION = enum.auto()
    PIP = enum.auto()


class AccountForm(QtWidgets.QHBoxLayout):
    """Group of controls for opening external items."""

    def __init__(self) -> None:
        """Initialize new :class:`~AccountForm` instance."""
        super().__init__()

        self.__edit: typing.Final[widgets.LineEditBase] = widgets.LineEditBase()
        self.__edit.setReadOnly(True)

        self.__button: typing.Final[common.OpenIconButton] = common.OpenIconButton()

        self.addWidget(self.__edit)
        self.addWidget(self.__button)
        self.setSpacing(8)
        self.setContentsMargins(0, 8, 0, 0)

    @property
    def button(self) -> common.OpenIconButton:  # noqa: D401
        """Button for "open" action."""
        return self.__button

    @property
    def edit(self) -> widgets.LineEditBase:  # noqa: D401
        """Text control for selected value."""
        return self.__edit

    @property
    def group(self) -> utilities.WidgetGroup:  # noqa: D401
        """Group with dialog controls."""
        return utilities.WidgetGroup(self.edit, self.button)


class CondaDetails:
    """
    Environment-related details, retrieved from Conda.

    It should be parsed from `conda_data` results.
    """

    __slots__ = ('__environment_directories', '__environments')

    def __init__(self, source: typing.Mapping[str, typing.Any]) -> None:
        """Initialize new :class:`~CondaDetails` instance."""
        if not source:
            raise TypeError()
        try:
            self.__environment_directories: typing.Final[typing.Sequence[str]] = (
                source['processed_info']['__envs_dirs_writable']
            )
            self.__environments: typing.Final[typing.Mapping[str, str]] = {
                key: value
                for value, key in source['processed_info']['__environments'].items()
            }
        except (TypeError, ValueError, LookupError):
            raise ValueError() from None
        if not self.__environment_directories:
            raise ValueError()

    @property
    def environment_directories(self) -> typing.Sequence[str]:  # noqa: D401
        """List of directories, where new environments might be placed."""
        return self.__environment_directories

    @property
    def environments(self) -> typing.Mapping[str, str]:  # noqa: D401
        """Mapping of environment names to their prefixes."""
        return self.__environments


class ImportSelectorDialog(  # pylint: disable=too-many-instance-attributes
    multiaccount_dialogs.PrepopulatedSelectorDialog
):
    """Dialog for selecting the target to which to export an environment."""

    def __init__(self, parent: 'main_window.MainWindow') -> None:
        """Initialize new :class:`~BackupSelectorDialog` instance."""
        self.__conda_details: typing.Optional[CondaDetails] = None
        conda_worker = parent.api.conda_data()
        conda_worker.sig_chain_finished.connect(self.__parse_conda)

        self.__cloud_content: typing.Optional[typing.Mapping[str, typing.Any]] = None
        self.__cloud_controls: utilities.WidgetGroup = utilities.WidgetGroup()

        self.__environment_specification: typing.Optional[str] = None
        self.__environment_temporary: bool = False

        super().__init__(parent=parent)

        self.setWindowTitle('Import Environment')

    def __init_header__(  # pylint: disable=useless-super-delegation
            self,
            layout: QtWidgets.QVBoxLayout,
            *,
            caption_text: str = 'Import from:',
    ) -> None:
        """Initialize header part of the dialog."""
        super().__init_header__(layout, caption_text=caption_text)

    def __init_local_form__(self, form: multiaccount_dialogs.SelectorForm) -> None:
        """Initialize additional controls for local option."""
        content: typing.Final[AccountForm] = AccountForm()

        self.__local_environment: widgets.LineEditBase = content.edit
        self.__local_environment_type: ImportSelectorType = ImportSelectorType.NONE
        content.edit.textChanged.connect(self._process_local_environment)
        content.button.clicked.connect(self._process_open_local)

        form.add_layout(content, 1)
        self._controls.all += content.group

    def __init_cloud_form__(self, form: multiaccount_dialogs.SelectorForm) -> None:
        """Initialize additional controls for Cloud option."""
        super().__init_cloud_form__(form)

        content: typing.Final[AccountForm] = AccountForm()

        self.__cloud_environment: widgets.LineEditBase = content.edit
        content.edit.textChanged.connect(self._process_cloud_environment)
        content.button.clicked.connect(self._process_open_cloud)

        form.add_layout(content, 1)
        self._controls.all += content.group
        self.__cloud_controls = content.group

        self.__cloud_controls.disable()

    def __refresh_cloud_form__(self, form: multiaccount_dialogs.SelectorForm) -> None:
        """Refresh additional controls for Cloud option linked to log in status."""
        super().__refresh_cloud_form__(form)

        if not cloud.CloudAPI().username:
            self.__cloud_controls.disable()
            return

        worker: workers.TaskWorker = cloud.CloudAPI().list_environments.worker(  # pylint: disable=no-member
            limit=ENVIRONMENT_FETCH_LIMIT,
            offset=0,
            sort=ENVIRONMENT_ORDER,
        )
        self.finished.connect(worker.cancel)
        worker.signals.sig_done.connect(lambda result: self.finished.disconnect(worker.cancel))
        worker.signals.sig_done.connect(self.__check_cloud_fetch)
        worker.start()

    def __init_footer__(  # pylint: disable=useless-super-delegation
            self,
            layout: QtWidgets.QVBoxLayout,
            *,
            caption_text: str = 'New environment name:',
    ) -> None:
        """Initialize footer part of the dialog."""
        super().__init_footer__(layout, caption_text=caption_text)

    def __init_actions__(  # pylint: disable=useless-super-delegation
            self,
            layout: QtWidgets.QVBoxLayout,
            *,
            accept_text: str = 'Import',
            reject_text: str = 'Cancel',
    ) -> None:
        """Initialize actions part of the dialog."""
        super().__init_actions__(layout, accept_text=accept_text, reject_text=reject_text)

    @property
    def conda_details(self) -> typing.Optional[CondaDetails]:  # noqa: D401
        """Environment details retrieved from Conda."""
        return self.__conda_details

    @property
    def environment_specification(self) -> typing.Optional[str]:  # noqa: D401
        """Path to environment specification file, which should be used to create new environment."""
        return self.__environment_specification

    @property
    def environment_temporary(self) -> bool:  # noqa: D401
        """Is `environment_specification` is a temporary file and should be removed after creating new environment."""
        return self.__environment_temporary

    @property
    def local_environment(self) -> str:  # noqa: D401
        """Selected local environment value."""
        return self.__local_environment.text()

    @local_environment.setter
    def local_environment(self, value: str) -> None:
        """Update `local_environment` value."""
        self.__local_environment.setText(value)

    @property
    def local_environment_type(self) -> ImportSelectorType:  # noqa: D401
        """Type of the selected `local_environment`."""
        return self.__local_environment_type

    @local_environment_type.setter
    def local_environment_type(self, value: ImportSelectorType) -> None:
        """Update `local_environment_type` value."""
        self.__local_environment_type = value

    @property
    def cloud_environment(self) -> str:  # noqa: D401
        """Selected Cloud environment value."""
        return self.__cloud_environment.text()

    @cloud_environment.setter
    def cloud_environment(self, value: str) -> None:
        """Update `cloud_environment` value."""
        self.__cloud_environment.setText(value)

    def __parse_conda(self, worker: typing.Any, output: typing.Any, error: typing.Any) -> None:
        """Parse response of the conda data command."""
        try:
            self.__conda_details = CondaDetails(source=output)
        except (TypeError, ValueError):
            pass
        else:
            self.__update_acceptable()

    def _process_accept(self) -> None:
        """Process clicking on the 'OK' button."""
        if self.conda_details is None:
            raise ValueError()
        if (not self.environment_override) and (self.environment_name in self.conda_details.environments):
            self.footer_error = multiaccount_dialogs.FOOTER_ERROR_TEMPLATE.format(
                content='Please rename environment to continue.',
            )
            return

        self.set_busy(True)
        self.clear_heading_errors()
        self.footer_error = ''

        if self.selection == multiaccount_dialogs.SelectorValue.LOCAL:
            self.__process_accept_local()
        elif self.selection == multiaccount_dialogs.SelectorValue.CLOUD:
            self.__process_accept_cloud()
        else:
            raise NotImplementedError()

    def _process_selection(self, value: multiaccount_dialogs.SelectorValue) -> None:
        """Process changing selected value in the dialog."""
        super()._process_selection(value)
        self.__update_acceptable()

    def _process_environment_name(self, value: str) -> None:
        """Process change of the `environment_name` value."""
        super()._process_environment_name(value)
        self.__update_acceptable()

    def _process_local_environment(self, value: str) -> None:
        """Process change of the `local_environment` value."""
        self.__update_acceptable()

    def _process_cloud_environment(self, value: str) -> None:
        """Process change of the `cloud_environment` value."""
        self.__update_acceptable()

    def __update_acceptable(self) -> None:
        """Update state of the accept button according to current dialog state."""
        if (self.__conda_details is not None) and self.environment_name:
            if (self.selection == multiaccount_dialogs.SelectorValue.LOCAL) and self.local_environment:
                self.set_acceptable(True)
                return

            if (self.selection == multiaccount_dialogs.SelectorValue.CLOUD) and self.cloud_environment:
                self.set_acceptable(True)
                return

        self.set_acceptable(False)

    # Local

    def _process_open_local(self) -> None:
        """Open dialog to select local environment for import."""
        environment_files: typing.Final[str] = 'Conda environment files (*.yaml *.yml)'
        specification_files: typing.Final[str] = 'Conda explicit specification files (*.txt)'
        pip_files: typing.Final[str] = 'Pip requirement files (*.txt)'

        path: str
        selected_filter: str
        path, selected_filter = compat.getopenfilename(
            parent=self,
            caption='Import Environment',
            basedir=os.path.expanduser('~'),
            filters=';;'.join([environment_files, specification_files, pip_files]),
        )
        if not path:
            return

        self.value = multiaccount_dialogs.SelectorValue.LOCAL

        self.__local_environment.setText(path)

        if selected_filter == environment_files:
            self.__local_environment_type = ImportSelectorType.ENVIRONMENT
        elif selected_filter == specification_files:
            self.__local_environment_type = ImportSelectorType.SPECIFICATION
        elif selected_filter == pip_files:
            self.__local_environment_type = ImportSelectorType.PIP

        self.environment_name = os.path.splitext(os.path.basename(path))[0]

    def __process_accept_local(self) -> None:
        """Process clicking on the 'OK' button when local account is selected."""
        telemetry.ANALYTICS.instance.event('request-environment-import', {'from': 'local'})
        if self.__local_environment_type == ImportSelectorType.SPECIFICATION:
            self.__environment_specification = self.local_environment
            self.__environment_temporary = False
            telemetry.ANALYTICS.instance.event('import-environment', {'from': 'local'})
            super()._process_accept()
            return

        path: str = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex + '.yml')

        if self.__local_environment_type == ImportSelectorType.ENVIRONMENT:
            env_to_spec(source=self.local_environment, target=path, name=self.environment_name)

        if self.__local_environment_type == ImportSelectorType.PIP:
            pip_to_spec(source=self.local_environment, target=path, name=self.environment_name)

        self.__environment_specification = path
        self.__environment_temporary = True
        telemetry.ANALYTICS.instance.event('import-environment', {'from': 'local'})
        super()._process_accept()
        return

    # Cloud

    def __check_cloud_fetch(self, result: workers.TaskResult) -> None:
        """Parse available Cloud environments."""
        if result.status != workers.TaskStatus.SUCCEEDED:
            return

        try:
            self.__cloud_content = typing.cast(typing.Mapping[str, typing.Any], result.result)
        except workers.TaskCanceledError:
            self.__cloud_content = {'total': 0}

        total: int = self.__cloud_content['total']
        username: str = cloud.CloudAPI().username

        if total <= 0:
            self.cloud_account = (
                f'<span style="font-size: 12px; font-weight: 500">'
                f'No environments available for {html.escape(username)}'
                f'</span>'
            )
            return

        if total == 1:
            self.cloud_account = (
                f'<span style="font-size: 12px; font-weight: 500">'
                f'1 environment available for {html.escape(username)}'
                f'</span>'
            )
        else:
            self.cloud_account = (
                f'<span style="font-size: 12px; font-weight: 500">'
                f'{total} environments available for {html.escape(username)}'
                f'</span>'
            )
        self.__cloud_controls.enable()

    def _process_open_cloud(self) -> None:
        """Open dialog to select Cloud environment for import."""
        dialog: EnvSelectorDialog = EnvSelectorDialog(parent=self, envs=self.__cloud_content)
        if dialog.exec_():
            self.value = multiaccount_dialogs.SelectorValue.CLOUD
            self.environment_name = self.cloud_environment = dialog.current_item.name

    def __process_accept_cloud(self) -> None:
        """Process clicking on the 'OK' button when Cloud account is selected."""
        telemetry.ANALYTICS.instance.event('request-environment-import', {'from': 'cloud'})

        path: str = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex + '.yml')

        worker: workers.TaskWorker = cloud.CloudAPI().download_environment.worker(  # pylint: disable=no-member
            name=self.__cloud_environment.text(),
            path=path,
        )
        self.finished.connect(worker.cancel)
        worker.signals.sig_done.connect(lambda result: self.finished.disconnect(worker.cancel))
        worker.signals.sig_done.connect(self.__check_cloud_download)
        worker.start()

    def __check_cloud_download(self, result: workers.TaskResult) -> None:
        """Check results of a Cloud environment download."""
        if not self.isVisible():
            return

        if result.status == workers.TaskStatus.SUCCEEDED:
            self.__environment_specification = result.call.kwargs['path']
            self.__environment_temporary = True
            telemetry.ANALYTICS.instance.event('import-environment', {'from': 'cloud'})
            super()._process_accept()

        elif result.status == workers.TaskStatus.FAILED:
            handlers: cloud_tools.HttpErrorHandlers = cloud_tools.HttpErrorHandlers()
            handlers.register_handler(BaseException, self._handle_header_error(DEFAULT_ERROR))
            handlers.handle(exception=typing.cast(BaseException, result.exception))

        self.set_busy(False)


class EnvSelectorItem(environments.BaseListItemEnv):  # pylint: disable=missing-class-docstring
    env_item_height = styles.SASS_VARIABLES.WIDGET_IMPORT_ENVIRONMENT_TOTAL_HEIGHT


class EnvSelectorDialog(dialogs.DialogBase):
    """List environments dialog."""

    sig_env_list_ready = QtCore.Signal(object)

    def __init__(
            self,
            parent: typing.Optional[QtWidgets.QWidget] = None,
            envs: typing.Optional[typing.Mapping[str, typing.Any]] = None,
    ) -> None:
        """List environments dialog."""
        super().__init__(parent=parent)

        self.api: typing.Final['cloud._CloudAPI'] = cloud.CloudAPI()
        self.offset: int = 0

        # Widgets
        heading_label: typing.Final[QtWidgets.QLabel] = widgets.LabelBase()
        heading_label.setText('<span style="font-size: 16px; font-weight: 500">Select environment from Cloud:</span>')

        self.list: typing.Final[environments.BaseListWidgetEnv] = environments.BaseListWidgetEnv()
        self.list.setFocus()

        self.error_label: typing.Final[global_common.WarningLabel] = global_common.WarningLabel()

        progress_frame: typing.Final[multiaccount_dialogs.ProgressFrame] = (
            multiaccount_dialogs.ProgressFrame()
        )
        self.progress_bar: typing.Final[QtWidgets.QProgressBar] = progress_frame.progress_bar

        self.button_cancel: typing.Final[widgets.ButtonNormal] = widgets.ButtonNormal('Cancel')

        self.button_ok: typing.Final[widgets.ButtonPrimary] = widgets.ButtonPrimary('Select')
        self.button_ok.setDisabled(True)

        # Layouts
        layout_buttons: typing.Final[QtWidgets.QHBoxLayout] = QtWidgets.QHBoxLayout()
        layout_buttons.addWidget(progress_frame, 1)
        layout_buttons.addWidget(self.button_cancel)
        layout_buttons.addWidget(self.button_ok)
        layout_buttons.setContentsMargins(0, 0, 0, 0)
        layout_buttons.setSpacing(12)

        layout: typing.Final[QtWidgets.QVBoxLayout] = QtWidgets.QVBoxLayout()
        layout.addWidget(heading_label)
        layout.addWidget(widgets.SpacerVertical())
        layout.addWidget(self.list)
        layout.addWidget(widgets.SpacerVertical())
        layout.addWidget(self.error_label)
        layout.addWidget(widgets.SpacerVertical())
        layout.addLayout(layout_buttons)

        # Setup
        self.setLayout(layout)
        self.setMinimumHeight(475)
        self.setMinimumWidth(460)
        self.setWindowTitle('Import New Environment')
        self.setFocus()

        # Signals
        self.sig_env_list_ready.connect(self.extend_list)
        self.button_ok.clicked.connect(self.accept)
        self.button_cancel.clicked.connect(self.reject)
        self.list.verticalScrollBar().valueChanged.connect(self.__extend_list_by_scroll)

        if envs:
            self.extend_list(envs)
        else:
            self._load_environments()

    @property
    def current_item(self) -> EnvSelectorItem:  # noqa: D401
        """Current selected item."""
        return self.list.currentItem()

    @staticmethod
    def get_env_names(envs: typing.Mapping[str, typing.Any]) -> typing.List[str]:
        """Get list of environment names."""
        try:
            return [
                env['name']
                for env in envs['items']
            ]
        except (TypeError, KeyError):
            return []

    def __extend_list_by_scroll(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Download batch of environments and add their names to the list if scroll bar reached the bottom."""
        scroll_bar: QtWidgets.QScrollBar = self.list.verticalScrollBar()
        if scroll_bar.value() >= scroll_bar.maximum() - 3:
            self._load_environments()

    def __process_load_result(self, output: workers.TaskResult) -> None:
        """Check the output from worker and emit signal to extend list of environment names."""
        # if not self.isVisible():
        #     return

        self.progress_bar.setVisible(False)

        try:
            self.sig_env_list_ready.emit(output.result)
        except requests.RequestException:
            self.error_label.text = 'Unable to fetch available environments.'
        except workers.TaskCanceledError:
            pass
        else:
            if output.result.get('items'):
                self.list.verticalScrollBar().valueChanged.connect(self.__extend_list_by_scroll)

    def _load_environments(self) -> None:
        """Start loading a batch of environments in separate tread"""
        self.progress_bar.setVisible(True)
        self.list.verticalScrollBar().valueChanged.disconnect(self.__extend_list_by_scroll)

        worker: workers.TaskWorker = self.api.list_environments.worker(  # pylint: disable=no-member
            limit=ENVIRONMENT_FETCH_LIMIT,
            offset=self.offset,
            sort=ENVIRONMENT_ORDER,
        )
        worker.signals.sig_done.connect(self.__process_load_result)
        worker.start()

    def extend_list(self, envs: typing.Mapping[str, typing.Any]) -> None:
        """Add environment names to the list."""
        env_names: typing.Sequence[str] = self.get_env_names(envs)
        for name in env_names:
            self.list.addItem(EnvSelectorItem(name))

        self.offset += len(env_names)
        self.button_ok.setEnabled(self.offset > 0)



# -*- coding: utf-8 -*-

"""Additional utility functions to use with application detectors."""

from __future__ import annotations

__all__ = [
    'is_32bit', 'is_64bit', 'is_linux', 'is_not_linux', 'is_osx', 'is_not_osx', 'is_win', 'is_not_win',
    'CliVersionChecker', 'FileVersionChecker', 'InjectVersion',
]

import io
import os
import re
import typing

from anaconda_navigator import config as navigator_config
from anaconda_navigator.utils.conda import launch as conda_launch_utils

from .. import detectors
from ..detectors import utilities
from .. import validation


def is_32bit() -> bool:
    """Check if user runs a 32-bit OS."""
    return navigator_config.BITS_32


def is_64bit() -> bool:
    """Check if user runs a 64-bit OS."""
    return navigator_config.BITS_64


def is_linux() -> bool:
    """Check if user runs Linux OS."""
    return navigator_config.LINUX


def is_not_linux() -> bool:
    """Check if user runs anything but Linux OS."""
    return not navigator_config.LINUX


def is_osx() -> bool:
    """Check if user runs OS X."""
    return navigator_config.MAC


def is_not_osx() -> bool:
    """Check if user runs anything but OS X."""
    return not navigator_config.MAC


def is_win() -> bool:
    """Check if user runs Windows OS."""
    return navigator_config.WIN


def is_not_win() -> bool:
    """Check if user runs anything but Windows OS."""
    return not navigator_config.WIN


class CliVersionChecker(detectors.Filter):  # pylint: disable=too-few-public-methods
    """
    Detect version of the application from the terminal output of the application.

    It executes application with a given set of `arguments` and searches for `pattern` in output. The latter one must
    have a single group, value from which will be used as a result version.
    """

    __slots__ = ('__arguments', '__pattern')

    def __init__(
            self,
            pattern: str = r'(\d+\.[\d\.]+)',
            arguments: typing.Union[str, typing.Iterable[str]] = ('--version',),
    ) -> None:
        """Initialize new :class:`~CliVersionChecker` instance."""
        if isinstance(arguments, str):
            arguments = (arguments,)
        elif not isinstance(arguments, tuple):
            arguments = tuple(arguments)

        self.__arguments: typing.Final[typing.Tuple[str, ...]] = arguments
        self.__pattern: typing.Final[typing.Pattern[str]] = re.compile(pattern)

    @classmethod
    def _parse_configuration(cls, *args: typing.Any, **kwargs: typing.Any) -> detectors.Detector:
        """Parse configuration for this particular :class:`~Detector`."""
        validation.OfLength(0)(args, field='args')
        pattern: str = validation.ItemPopper('pattern', default=r'(\d+\.[\d\.]+)').convert(validation.IS_STR)(kwargs)
        arguments: typing.Tuple[str, ...] = validation.ItemPopper('arguments', default=('--version',)).convert(
            validation.sequence_of(str),
        )(kwargs)
        validation.IsEmptyMapping()(kwargs)
        with validation.catch_exception():
            return cls(pattern=pattern, arguments=arguments)

    def __call__(
            self,
            parent: typing.Iterator[detectors.DetectedApplication],
            *,
            context: detectors.DetectorContext,
    ) -> typing.Iterator[detectors.DetectedApplication]:
        """Iterate through detected applications."""
        application: 'detectors.DetectedApplication'
        for application in parent:
            if application.has_executable:
                stdout: str
                try:
                    stdout, _, _ = conda_launch_utils.run_process([application.executable, *self.__arguments])
                    match: typing.Optional[typing.Match[str]] = self.__pattern.search(stdout)
                    if match:
                        application = application.replace(version=match.group(1))
                except (OSError, LookupError, TypeError, ValueError):
                    pass
            yield application


class FileVersionChecker(detectors.Filter):  # pylint: disable=too-few-public-methods
    """Detect version of the application from the file in the application :code:`root`."""

    __slots__ = ('__encoding', '__filename', '__pattern')

    def __init__(self, filename: str, pattern: str = r'(\d+\.[\d\.]+)', encoding: str = 'utf8') -> None:
        """Initialize new :class:`~FileVersionChecker` instance."""
        self.__encoding: typing.Final[str] = encoding
        self.__filename: typing.Final[str] = filename
        self.__pattern: typing.Final[typing.Pattern[str]] = re.compile(pattern)

    @classmethod
    def _parse_configuration(cls, *args: typing.Any, **kwargs: typing.Any) -> detectors.Detector:
        """Parse configuration for this particular :class:`~Detector`."""
        validation.OfLength(0)(args, field='args')
        encoding: str = validation.ItemPopper('encoding', default='utf8').convert(validation.IS_STR)(kwargs)
        filename: str = utilities.parse_and_join(validation.ItemPopper('filename')(kwargs))  # type: ignore
        pattern: str = validation.ItemPopper('pattern', default=r'(\d+\.[\d\.]+)').convert(validation.IS_STR)(kwargs)
        validation.IsEmptyMapping()(kwargs)
        with validation.catch_exception():
            return cls(encoding=encoding, filename=filename, pattern=pattern)

    def __call__(
            self,
            parent: typing.Iterator[detectors.DetectedApplication],
            *,
            context: detectors.DetectorContext,
    ) -> typing.Iterator[detectors.DetectedApplication]:
        """Iterate through detected applications."""
        application: 'detectors.DetectedApplication'
        for application in parent:
            if application.has_root:
                try:
                    stream: io.TextIOWrapper
                    with open(
                            os.path.join(application.root, self.__filename),
                            'rt',
                            encoding=self.__encoding,
                    ) as stream:
                        match: typing.Optional[typing.Match[str]] = self.__pattern.search(stream.read(4_194_304))
                        if match:
                            application = application.replace(version=match.group(1))
                except (OSError, LookupError, TypeError, ValueError):
                    pass
            yield application


class InjectVersion(detectors.Filter):  # pylint: disable=too-few-public-methods
    """Register a predefined version for the application."""

    __slots__ = ('__force', '__value')

    def __init__(self, value: str, *, force: bool = False) -> None:
        """Initialize new :class:`~InjectVersion` instance."""
        self.__force: typing.Final[bool] = force
        self.__value: typing.Final[str] = str(value).strip()

    @classmethod
    def _parse_configuration(cls, *args: typing.Any, **kwargs: typing.Any) -> detectors.Detector:
        """Parse configuration for this particular :class:`~Detector`."""
        validation.OfLength(0)(args, field='args')
        force = validation.ItemPopper('force', default=False).convert(validation.IS_BOOL)(kwargs)
        value = validation.ItemPopper('value').convert(validation.IS_STR)(kwargs)
        validation.IsEmptyMapping()(kwargs)
        with validation.catch_exception():
            return cls(force=force, value=value)

    def __call__(
            self,
            parent: typing.Iterator[detectors.DetectedApplication],
            *,
            context: detectors.DetectorContext,
    ) -> typing.Iterator[detectors.DetectedApplication]:
        """Iterate through detected applications."""
        application: 'detectors.DetectedApplication'
        for application in parent:
            if self.__force or (not application.has_version):
                application = application.replace(version=self.__value)
            yield application



# -*- coding: utf8 -*-

"""Simple provider to store telemetry into a file."""

from __future__ import annotations

__all__ = ['FileProvider']

import json

import collections.abc
import typing

from . import core


class FileProvider(core.Provider):
    """
    Provider that writes all telemetry to the file.

    :param file: Path to a file to write telemetry to.
    """

    __slots__ = ('__stream',)

    def __init__(self, context: core.Context, file: str, *, alias: str = 'file') -> None:
        """Initialize new instance of a :class:`~FileProvider`."""
        super().__init__(alias=alias, context=context)

        # pylint: disable=consider-using-with
        self.__stream: typing.Final[typing.TextIO] = open(file, 'at', encoding='utf8')

    def close(self) -> None:
        """Close provider."""
        self.__stream.close()

    def process(self, events: collections.abc.MutableSequence[core.Event]) -> None:
        """Process all pending events."""
        for event in events:
            json.dump(
                {
                    'key': event.key,
                    'name': event.name,
                    'properties': event.properties,
                    'timestamp': event.timestamp.isoformat(),
                    'user_properties': event.user_properties,
                },
                self.__stream,
            )
            self.__stream.write('\n')
        events.clear()
